import copy
import math
import tempfile
import time

import numpy as np
from typing import Optional

import pandas as pd
from matplotlib import pyplot as plt
from scipy.constants import golden
from scipy.special import logsumexp
from bilby.core.prior import Constraint, DeltaFunction

from dingo.core.dataset import DingoDataset
from dingo.core.density import train_unconditional_density_estimator
from dingo.core.utils.misc import recursive_check_dicts_are_equal
from dingo.core.utils.plotting import plot_corner_multi

DATA_KEYS = [
    "samples",
    "context",
    "event_metadata",
    "importance_sampling_metadata",
    "log_evidence",
    "log_noise_evidence",
]


class Result(DingoDataset):
    """
    A dataset class to hold a collection of samples, implementing I/O, importance
    sampling, and unconditional flow training.

    Attributes:
        samples : pd.Dataframe
            Contains parameter samples, as well as (possibly) log_prob, log_likelihood,
            weights, log_prior, delta_log_prob_target.
        domain : Domain
            Should be implemented in a subclass.
        prior : PriorDict
            Should be implemented in a subclass.
        likelihood : Likelihood
            Should be implemented in a subclass.
        context : dict
            Context data from which the samples were produced (e.g., strain data, ASDs).
        metadata : dict
        event_metadata : dict
        log_evidence : float
        log_evidence_std : float (property)
        effective_sample_size, n_eff : float (property)
        sample_efficiency : float (property)
    """

    dataset_type = "core_result"

    def __init__(self, file_name=None, dictionary=None):
        self.event_metadata = None
        self.context = None
        self.samples = None
        self.log_noise_evidence = None
        super().__init__(
            file_name=file_name,
            dictionary=dictionary,
            data_keys=DATA_KEYS,
        )

        # Initialize as empty dict, so we can fill it up later.
        if self.importance_sampling_metadata is None:
            self.importance_sampling_metadata = {}

        self._build_prior()
        self._build_domain()
        if self.importance_sampling_metadata.get("updates"):
            self._rebuild_domain()

    @property
    def metadata(self):
        return self.settings

    @property
    def base_metadata(self):
        if self.metadata["train_settings"]["data"].get("unconditional", False):
            return self.metadata["base"]
        else:
            return self.metadata

    @property
    def injection_parameters(self):
        if self.context:
            return self.context.get("parameters")
        else:
            return None

    @property
    def constraint_parameter_keys(self):
        return [k for k, v in self.prior.items() if isinstance(v, Constraint)]

    @property
    def search_parameter_keys(self):
        return [
            k
            for k, v in self.prior.items()
            if (not isinstance(v, Constraint) and not isinstance(v, DeltaFunction))
        ]

    @property
    def fixed_parameter_keys(self):
        return [k for k, v in self.prior.items() if isinstance(v, DeltaFunction)]

    def _build_domain(self):
        self.domain = None

    def _build_prior(self):
        self.prior = None

    def _build_likelihood(self, **likelihood_kwargs):
        self.likelihood = None

    def reset_event(self, event_dataset):
        """
        Set the Result context and event_metadata based on an EventDataset.

        If these attributes already exist, perform a comparison to check for changes.
        Update relevant objects appropriately. Note that setting context and
        event_metadata attributes directly would not perform these additional checks and
        updates.

        Parameters
        ----------
        event_dataset: EventDataset
            New event to be used for importance sampling.
        """
        context = event_dataset.data
        event_metadata = event_dataset.settings

        if self.context is not None and not check_equal_dict_of_arrays(
            self.context, context
        ):
            # This is really just for notification. Actions are only taken if the
            # event metadata differ.
            print("\nNew event data differ from existing.")
        self.context = context

        if self.event_metadata is not None and self.event_metadata != event_metadata:
            print("Changes")
            print("=======")
            old_minus_new = dict(freeze(self.event_metadata) - freeze(event_metadata))
            print("Old event metadata:")
            for k in sorted(old_minus_new):
                print(f"  {k}:  {self.event_metadata[k]}")

            new_minus_old = dict(freeze(event_metadata) - freeze(self.event_metadata))
            print("New event metadata:")
            if self.importance_sampling_metadata.get("updates") is None:
                self.importance_sampling_metadata["updates"] = {}
            for k in sorted(new_minus_old):
                print(f"  {k}:  {event_metadata[k]}")
                self.importance_sampling_metadata["updates"][k] = event_metadata[k]

            self._rebuild_domain(verbose=True)
        self.event_metadata = event_metadata

    def _rebuild_domain(self, verbose=False):
        pass

    @property
    def num_samples(self):
        if self.samples is not None:
            return len(self.samples)
        else:
            return 0

    @property
    def effective_sample_size(self):
        if "weights" in self.samples:
            weights = self.samples["weights"]
            return np.sum(weights) ** 2 / np.sum(weights**2)
        else:
            return None

    @property
    def n_eff(self):
        return self.effective_sample_size

    @property
    def sample_efficiency(self):
        if "weights" in self.samples:
            return self.effective_sample_size / len(self.samples)
        else:
            return None

    @property
    def log_evidence_std(self):
        if "weights" in self.samples and self.log_evidence:
            return np.sqrt(
                (self.num_samples - self.n_eff) / (self.num_samples * self.n_eff)
            )
        else:
            return None

    @property
    def log_bayes_factor(self):
        if self.log_evidence and self.log_noise_evidence:
            return self.log_evidence - self.log_noise_evidence
        else:
            return None

    def importance_sample(self, num_processes: int = 1, **likelihood_kwargs):
        """
        Calculate importance weights for samples.

        Importance sampling starts with samples have been generated from a proposal
        distribution q(theta), in this case a neural network model. Certain networks
        (i.e., non-GNPE) also provide the log probability of each sample,
        which is required for importance sampling.

        Given the proposal, we re-weight samples according to the (un-normalized)
        target distribution, which we take to be the likelihood L(theta) times the
        prior pi(theta). This gives sample weights

            w(theta) ~ pi(theta) L(theta) / q(theta),

        where the overall normalization does not matter (and we take to have mean 1).
        Since q(theta) enters this expression, importance sampling is only possible
        when we know the log probability of each sample.

        As byproducts, this method also estimates the evidence and effective sample
        size of the importance sampled points.

        This method modifies the samples pd.DataFrame in-place, adding new columns for
        log_likelihood, log_prior, and weights. It also stores the log_evidence as an
        attribute.

        Parameters
        ----------
        num_processes : int
            Number of parallel processes to use when calculating likelihoods. (This is
            the most expensive task.)
        likelihood_kwargs : dict
            kwargs that are forwarded to the likelihood constructor. E.g., options for
            marginalization.
        """

        if self.samples is None:
            raise KeyError("Proposal samples are required for importance sampling.")
        if "log_prob" not in self.samples:
            raise KeyError(
                "Stored samples do not include log probability, which is "
                "needed for importance sampling. To obtain the log probability, "
                "it is necessary to train an unconditional flow based on the existing "
                "samples. This can then be sampled with log probability."
            )

        self._build_likelihood(**likelihood_kwargs)

        if "delta_log_prob_target" in self.samples.columns:
            delta_log_prob_target = self.samples["delta_log_prob_target"].to_numpy()
        else:
            delta_log_prob_target = 0.0

        # select parameters in self.samples (required as log_prob and potentially gnpe
        # proxies are also stored in self.samples, but are not needed for the likelihood.
        # TODO: replace by self.metadata["train_settings"]["data"]["inference_parameters"]
        param_keys = [k for k, v in self.prior.items() if not isinstance(v, Constraint)]
        theta = self.samples[param_keys]

        # Calculate the (un-normalized) target density as prior times likelihood,
        # evaluated at the same sample points.
        log_prior = self.prior.ln_prob(theta, axis=0)

        # The prior or delta_log_prob_target may be -inf for certain samples.
        # For these, we do not want to evaluate the likelihood, in particular because
        # it may not even be possible to generate signals outside the prior (e.g.,
        # for BH spins > 1).
        valid_samples = (log_prior + delta_log_prob_target) != -np.inf
        theta = theta.iloc[valid_samples]

        print(f"Calculating {len(theta)} likelihoods.")
        t0 = time.time()
        log_likelihood = self.likelihood.log_likelihood_multi(
            theta, num_processes=num_processes
        )
        print(f"Done. This took {time.time() - t0:.2f} seconds.")

        self.log_noise_evidence = self.likelihood.log_Zn
        self.samples["log_prior"] = log_prior
        self.samples.loc[valid_samples, "log_likelihood"] = log_likelihood
        self._calculate_evidence()

    def _calculate_evidence(self):
        """Calculate the Bayesian log evidence and sample weights.

        This is called at the end of importance sampling, when changing the prior,
        and when combining Results.

        The evidence

            Z = \\int d\\theta \\pi(\\theta) L(\\theta),

        where \\pi = prior, L = likelihood.

        For importance sampling, we estimate this using Monte Carlo integration using
        the proposal distribution q(\\theta),

            Z = \\int d\\theta q(\\theta) \\pi(\\theta) L(\\theta) / q(\\theta)
            \\sim (1/N) \\sum_i \\pi(\\theta_i) L(\\theta_i) / q(\\theta_i)

        where we are summing over samples \\theta_i \\sim q(\\theta).

        The integrand is just the importance weight (prior to any normalization). It
        is more numerically stable to evaluate \\log(Z),

            \\log Z \\sim \\log \\sum_i \\exp( \\log \\pi_i + \\log L_i - \\log q_i ) -
            \\log N
            = logsumexp ( log_weights ) - log N

        Notes
        -----
        * We use the logsumexp function, which is more numerically stable.
        * N = num_samples is the total number of samples (including the
            zero-weight samples).
        * q, \\pi, L must be distributions in the same parameter space (the same
            coordinates). We have undone any standardizations so this is the case.
        """
        if (
            "log_prob" in self.samples
            and "log_likelihood" in self.samples
            and "log_prior" in self.samples
        ):
            log_prob_proposal = self.samples["log_prob"]
            log_prior = self.samples["log_prior"]
            log_likelihood = self.samples["log_likelihood"]
            if "delta_log_prob_target" in self.samples:
                delta_log_prob_target = self.samples["delta_log_prob_target"]
            else:
                delta_log_prob_target = 0.0

            # *Un-normalized* log weights are needed to calculate evidence.
            log_weights = (
                log_prior
                + np.nan_to_num(log_likelihood)  # NaN = no log_likelihood evaluation
                + delta_log_prob_target
                - np.nan_to_num(
                    log_prob_proposal
                )  # NaN = outside prior so no synthetic
                # phase
            )
            self.log_evidence = logsumexp(log_weights) - np.log(self.num_samples)

            # Save the *normalized* weights.
            weights = np.exp(log_weights - np.max(log_weights))
            weights /= np.mean(weights)
            self.samples["weights"] = weights

    def sampling_importance_resampling(self, num_samples=None, random_state=None):
        """
        Generate unweighted posterior samples from weighted ones. New
        samples are sampled with probability proportional to the sample weight.
        Resampling is done with replacement, until the desired number of
        unweighted samples is obtained.

        Parameters
        ----------
        num_samples : int
            Number of samples to resample.
        random_state : int or None
            Sampling seed.

        Returns
        -------
        pd.Dataframe
            Unweighted samples
        """
        if num_samples is None:
            num_samples = len(self.samples)

        if num_samples > len(self.samples):
            raise ValueError("Cannot sample more points than in the weighted posterior")

        unweighted_samples = self.samples.sample(
            n=num_samples,
            weights=self.samples["weights"],
            replace=True,
            ignore_index=True,
            random_state=random_state,
        )
        return unweighted_samples.drop(["weights"], axis=1)

    def parameter_subset(self, parameters):
        """
        Return a new object of the same type, with only a subset of parameters. Drops
        all other columns in samples DataFrame as well (e.g., log_prob, weights).

        Parameters
        ----------
        parameters : list
            List of parameters to keep.

        Returns
        -------
        Result
        """
        result_dict = self.to_dictionary()
        result_dict["samples"] = self.samples[
            parameters
        ]  # Drop log_probs, weights, etc.
        return type(self)(dictionary=result_dict)

    def train_unconditional_flow(
        self,
        parameters,
        nde_settings: dict,
        train_dir: Optional[str] = None,
        threshold_std: Optional[float] = np.inf,
    ):
        """
        Train an unconditional flow to represent the distribution of self.samples.

        Parameters
        ----------
        parameters : list
            List of parameters over which to train the flow. Can be a subset of the
            existing parameters.
        nde_settings : dict
            Configuration settings for the neural density estimator.
        train_dir : Optional[str]
            Where to save the output of network training, e.g., logs, checkpoints. If
            not provide, a temporary directory is used.
        threshold_std : Optional[float]
            Drop samples more than threshold_std standard deviations away from the mean
            (in any parameter) before training the flow. This is meant to remove outlier
            samples.

        Returns
        -------
        PosteriorModel
        """
        sub_result = self.parameter_subset(parameters)

        # Filter outliers, as they decrease the performance of the density estimator.
        mean = np.mean(sub_result.samples, axis=0)
        std = np.std(sub_result.samples, axis=0)
        lower, upper = mean - threshold_std * std, mean + threshold_std * std
        inds = np.where(
            np.all((lower <= sub_result.samples), axis=1)
            * np.all((sub_result.samples <= upper), axis=1)
        )[0]
        if len(inds) / len(sub_result.samples) < 0.95:
            raise ValueError("Too many proxy samples outside of specified range.")
        sub_result.samples = sub_result.samples.iloc[inds]
        nde_settings["data"] = {"inference_parameters": parameters}

        temporary_directory = None
        if train_dir is None:
            temporary_directory = tempfile.TemporaryDirectory()
            train_dir = temporary_directory.name

        unconditional_model = train_unconditional_density_estimator(
            sub_result,
            nde_settings,
            train_dir,
        )

        if temporary_directory is not None:
            temporary_directory.cleanup()

        # unconditional_model.save_model("temp_model.pt")
        return unconditional_model

        # Note: self.gnpe_proxy_sampler.transform_post, and self.transform_post *must*
        # contain the SelectStandardizeRepackageParameters transformation, such that
        # the log_prob is correctly de-standardized!

    def print_summary(self):
        """
        Display the number of samples, and (if importance sampling is complete) the log
        evidence and number of effective samples.
        """
        print("Number of samples:", len(self.samples))
        if self.log_evidence is not None:
            print(
                f"Log(evidence): {self.log_evidence:.3f} +- {self.log_evidence_std:.3f}"
            )
            print(
                f"Effective samples {self.n_eff:.1f}: "
                f"(Sample efficiency = {100 * self.sample_efficiency:.2f}%)"
            )

    def split(self, num_parts):
        """
        Split the Result into a set of smaller results. The samples are evenly divided
        among the sub-results. Additional information (metadata, context, etc.) are
        copied into each.

        This is useful for splitting expensive tasks such as importance sampling across
        multiple jobs.

        Parameters
        ----------
        num_parts : int
            The number of parts to split the Result across.

        Returns
        -------
        list of sub-Results.
        """

        # Prepare a dictionary of all contents except the samples.
        dataset_dict_template = self.to_dictionary()
        del dataset_dict_template["samples"]

        part_size = self.num_samples // num_parts
        parts = []
        for i in range(num_parts):
            part_dict = copy.deepcopy(dataset_dict_template)
            if i < num_parts - 1:
                samples = self.samples.iloc[i * part_size : (i + 1) * part_size].copy()
            else:
                samples = self.samples.iloc[i * part_size :].copy()
            samples.reset_index(drop=True, inplace=True)
            part_dict["samples"] = samples
            part = type(self)(dictionary=part_dict)

            # Re-calculate evidence since it will differ for the new set of samples.
            part._calculate_evidence()
            parts.append(part)

        return parts

    @classmethod
    def merge(cls, parts):
        """
        Merge several Result instances into one. Check that they are compatible,
        in the sense of having the same metadata. Finally, calculate a new log evidence
        for the combined result.

        This is useful when recombining separate importance sampling jobs.

        Parameters
        ----------
        parts : list[Result]
            List of sub-Results to be combined.

        Returns
        -------
        Combined Result.
        """
        dataset_dict = parts[0].to_dictionary()
        del dataset_dict["log_evidence"]
        samples_parts = [dataset_dict.pop("samples")]

        for part in parts[1:]:
            part_dict = part.to_dictionary()
            del part_dict["log_evidence"]
            samples_parts.append(part_dict.pop("samples"))

            # Make sure we are not merging incompatible results. We deleted the
            # log_evidence since this can differ among the sub-results. Note that this
            # will also raise an error if files were created with different versions of
            # dingo.
            if not recursive_check_dicts_are_equal(part_dict, dataset_dict):
                raise ValueError("Results to be merged must have same metadata.")

        dataset_dict["samples"] = pd.concat(samples_parts, ignore_index=True)
        merged_result = cls(dictionary=dataset_dict)

        # Re-calculate the evidence based on the entire sample set.
        merged_result._calculate_evidence()
        return merged_result

    #
    # Plotting
    #

    def _cleaned_samples(self):
        """Return samples that exclude -inf and nan. This is used primarily for
        plotting."""

        # Do not plot any samples with -inf or nan. -inf can occur in
        # delta_log_prob_target or log_prior. nan occurs in log_likelihood when
        # log_likelihood not actually evaluated due to -inf in other columns (i.e.,
        # out of prior).

        return self.samples.replace(-np.inf, np.nan).dropna(axis=0)

    def plot_corner(self, parameters=None, filename="corner.pdf"):
        """
        Generate a corner plot of the samples.

        Parameters
        ----------
        parameters : list[str]
            List of parameters to include. If None, include all parameters.
            (Default: None)
        filename : str
            Where to save samples.
        """
        theta = self._cleaned_samples()
        # delta_log_prob_target is not interesting so never plot it.
        theta = theta.drop(columns="delta_log_prob_target", errors="ignore")

        # User option to plot specific parameters.
        if parameters:
            theta = theta[parameters]

        if "weights" in theta:
            plot_corner_multi(
                [theta, theta],
                weights=[None, theta["weights"].to_numpy()],
                labels=["Dingo", "Dingo-IS"],
                filename=filename,
            )
        else:
            plot_corner_multi(
                theta,
                labels="Dingo",
                filename=filename,
            )

    def plot_log_probs(self, filename="log_probs.png"):
        """
        Make a scatter plot of the target versus proposal log probabilities. For the
        target, subtract off the log evidence.
        """
        theta = self._cleaned_samples()
        if "log_likelihood" in theta:
            log_prob_proposal = theta["log_prob"].to_numpy()
            if "delta_log_prob_target" in theta:
                log_prob_proposal -= theta["delta_log_prob_target"].to_numpy()

            log_prior = theta["log_prior"].to_numpy()
            log_likelihood = theta["log_likelihood"].to_numpy()

            x = log_prob_proposal
            y = log_prior + log_likelihood - self.log_evidence

            plt.figure(figsize=(6, 6))
            plt.xlabel("proposal log_prob")
            plt.ylabel("target log_prob - log_evidence")
            y_lower, y_upper = np.max(y) - 20, np.max(y)
            plt.ylim(y_lower, y_upper)
            n_below = len(np.where(y < y_lower)[0])
            plt.title(
                f"Target log probabilities\n({n_below} below {y_lower:.2f})\n"
                f"log(evidence) = {self.log_evidence:.3f} +- {self.log_evidence_std:.3f}"
            )
            plt.scatter(x, y, s=0.5)
            plt.plot([y_upper - 20, y_upper], [y_upper - 20, y_upper], color="black")
            plt.tight_layout()
            plt.savefig(filename)
        else:
            print("Results not importance sampled. Cannot produce log_prob plot.")

    def plot_weights(self, filename="weights.png"):
        """Make a scatter plot of samples weights vs log proposal."""
        theta = self._cleaned_samples()
        if "weights" in theta and "log_prob" in theta:
            x = theta["log_prob"].to_numpy()
            y = theta["weights"].to_numpy()
            y /= y.mean()

            plt.figure(figsize=(6 * golden, 6))
            plt.xlabel("proposal log_prob")
            plt.ylabel("weight (normalized)")
            y_lower = 1e-4
            y_upper = math.ceil(
                np.max(y) / 10 ** math.ceil(np.log10(np.max(y)) - 1)
            ) * 10 ** math.ceil(np.log10(np.max(y)) - 1)
            plt.ylim(y_lower, y_upper)
            n_below = len(np.where(y < y_lower)[0])
            plt.yscale("log")
            plt.title(
                f"Importance sampling weights\n({n_below} below {y_lower})\n"
                f"Effective samples: {self.n_eff:.0f} (Efficiency = "
                f"{100 * self.sample_efficiency:.2f}%)."
            )
            plt.scatter(x, y, s=0.5)
            plt.tight_layout()
            plt.savefig(filename)
        else:
            print("Results not importance sampled. Cannot plot weights.")


def check_equal_dict_of_arrays(a, b):

    if type(a) != type(b):
        return False

    if isinstance(a, dict):
        a_keys = set(a.keys())
        b_keys = set(b.keys())
        if a_keys != b_keys:
            return False

        for k in a_keys:
            if not check_equal_dict_of_arrays(a[k], b[k]):
                return False

        return True

    elif isinstance(a, np.ndarray):
        return np.array_equal(a, b)

    else:
        raise TypeError(f"Cannot compare items of type {type(a)}")


def freeze(d):
    if isinstance(d, dict):
        return frozenset((key, freeze(value)) for key, value in d.items())
    elif isinstance(d, list):
        return tuple(freeze(value) for value in d)
    return d



import os
import sys
from os.path import join, isfile
import yaml
import argparse

from dingo.gw.training import (
    prepare_training_new,
    prepare_training_resume,
    train_stages,
)


def create_submission_file(train_dir, condor_settings, filename="submission_file.sub"):
    """
    TODO: documentation
    :param train_dir:
    :param filename:
    :return:
    """
    lines = []
    lines.append(f'executable = {condor_settings["executable"]}\n')
    lines.append(f'request_cpus = {condor_settings["num_cpus"]}\n')
    lines.append(f'request_memory = {condor_settings["memory_cpus"]}\n')
    lines.append(f'request_gpus = {condor_settings["num_gpus"]}\n')
    lines.append(
        f"requirements = TARGET.CUDAGlobalMemoryMb > "
        f'{condor_settings["memory_gpus"]}\n\n'
    )
    lines.append(f'arguments = {condor_settings["arguments"]}\n')
    lines.append(f'error = {join(train_dir, "info.err")}\n')
    lines.append(f'output = {join(train_dir, "info.out")}\n')
    lines.append(f'log = {join(train_dir, "info.log")}\n')
    lines.append("queue")

    with open(join(train_dir, filename), "w") as f:
        for line in lines:
            f.write(line)


def copyfile(src, dst):
    os.system("cp -p %s %s" % (src, dst))


def copy_logfiles(log_dir, epoch, name="info", suffixes=(".err", ".log", ".out")):
    for suffix in suffixes:
        src = join(log_dir, name + suffix)
        dest = join(log_dir, name + "_{:03d}".format(epoch) + suffix)
        try:
            copyfile(src, dest)
        except:
            print("Could not copy " + src)


def train_condor():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--train_dir", required=True, help="Directory for Dingo training output."
    )
    parser.add_argument("--checkpoint", default="model_latest.pt")
    parser.add_argument("--start_submission", action="store_true")
    args = parser.parse_args()

    # For condor settings, first try looking for a local settings file. Otherwise,
    # defer to train_settings.yaml.
    # if isfile(join(args.train_dir, 'local_settings.yaml')):
    #     with open(join(args.train_dir, 'local_settings.yaml')) as f:
    #         condor_settings = yaml.safe_load(f)['condor']
    # else:

    if not args.start_submission:
        #
        # TRAIN
        #

        if not isfile(join(args.train_dir, args.checkpoint)):
            print("Beginning new training run.")
            with open(join(args.train_dir, "train_settings.yaml"), "r") as f:
                train_settings = yaml.safe_load(f)

            # Extract the local settings from train settings file, save it separately.
            # This file can later be modified, and the settings take effect immediately
            # upon resuming.

            local_settings = train_settings.pop("local")
            with open(os.path.join(args.train_dir, "local_settings.yaml"), "w") as f:
                if local_settings.get("wandb", False) and "id" not in local_settings["wandb"].keys():
                    try:
                        import wandb
                        local_settings["wandb"]["id"] = wandb.util.generate_id()
                    except ImportError:
                        print("wandb not installed, cannot generate run id.")
                yaml.dump(local_settings, f, default_flow_style=False, sort_keys=False)

            pm, wfd = prepare_training_new(
                train_settings, args.train_dir, local_settings
            )

        else:
            print("Resuming training run.")
            with open(os.path.join(args.train_dir, "local_settings.yaml"), "r") as f:
                local_settings = yaml.safe_load(f)
            pm, wfd = prepare_training_resume(
                join(args.train_dir, args.checkpoint),
                local_settings,
                train_dir=args.train_dir,
            )

        complete = train_stages(pm, wfd, args.train_dir, local_settings)

        print("Copying log files")
        copy_logfiles(args.train_dir, epoch=pm.epoch)

        #
        # PREPARE NEXT SUBMISSION
        #

        if complete:
            print("Training complete, job will not be resubmitted.")
            sys.exit()

        else:
            condor_arguments = f"--train_dir {args.train_dir}"

    else:
        #
        # PREPARE FIRST SUBMISSION
        #

        condor_arguments = f"--train_dir {args.train_dir}"
        if args.checkpoint != "model_latest.pt":
            condor_arguments += f" --checkpoint {args.checkpoint}"

    submission_file = "submission_file.sub"
    with open(join(args.train_dir, "train_settings.yaml"), "r") as f:
        condor_settings = yaml.safe_load(f)["local"]["condor"]
    condor_settings["arguments"] = condor_arguments
    condor_settings["executable"] = join(
        os.path.dirname(sys.executable), "dingo_train_condor"
    )
    create_submission_file(args.train_dir, condor_settings, submission_file)

    #
    # SUBMIT NEXT CONDOR JOB
    #

    # There was no 'bid' in the sample settings file.
    bid = condor_settings["bid"]
    os.system(f"condor_submit_bid {bid} " f"{join(args.train_dir, submission_file)}")


if __name__ == "__main__":
    train_condor()



from bilby_pipe.job_creation.nodes import PlotNode as BilbyPlotNode

from dingo.pipe.utils import _strip_unwanted_submission_keys


class PlotNode(BilbyPlotNode):
    def __init__(self, inputs, merged_node, dag):
        super(BilbyPlotNode, self).__init__(inputs)
        self.dag = dag
        self.job_name = merged_node.job_name.replace("_merge", "") + "_plot"
        self.label = merged_node.job_name.replace("_merge", "") + "_plot"
        self.request_cpus = 1
        self.setup_arguments(
            add_ini=False, add_unknown_args=False, add_command_line_args=False
        )
        self.arguments.add("label", self.label)
        self.arguments.add("result", merged_node.result_file)
        self.arguments.add("outdir", self.inputs.result_directory)
        for plot_type in ["corner", "weights", "log_probs"]:
            if getattr(inputs, f"plot_{plot_type}", False):
                self.arguments.add_flag(plot_type)
        # self.arguments.add("format", inputs.plot_format)
        self.process_node()
        self.job.add_parent(merged_node.job)

        if self.inputs.simple_submission:
            _strip_unwanted_submission_keys(self.job)

    @property
    def executable(self):
        return self._get_executable_path("dingo_pipe_plot")



import sys
from multiprocessing import Pool

import numpy as np
import pandas as pd
from scipy.fft import fft
from scipy.special import logsumexp
from bilby.gw.utils import ln_i0
from threadpoolctl import threadpool_limits

from dingo.core.likelihood import Likelihood
from dingo.gw.injection import GWSignal
from dingo.gw.waveform_generator import WaveformGenerator
from dingo.gw.domains import build_domain
from dingo.gw.data.data_preparation import get_event_data_and_domain


class StationaryGaussianGWLikelihood(GWSignal, Likelihood):
    """
    Implements GW likelihood for stationary, Gaussian noise.
    """

    def __init__(
        self,
        wfg_kwargs,
        wfg_domain,
        data_domain,
        event_data,
        t_ref=None,
        time_marginalization_kwargs=None,
        phase_marginalization_kwargs=None,
        calibration_marginalization_kwargs=None,
        phase_grid=None,
    ):
        # TODO: Does the phase_grid argument ever get used?
        """
        Parameters
        ----------
        wfg_kwargs: dict
            Waveform generator parameters (at least approximant and f_ref).
        wfg_domain : dingo.gw.domains.Domain
            Domain used for waveform generation. This can potentially deviate from the
            final domain, having a wider frequency range needed for waveform generation.
        data_domain: dingo.gw.domains.Domain
            Domain object for event data.
        event_data: dict
            GW data. Contains strain data in event_data["waveforms"] and asds in
            event_data["asds"].
        t_ref: float
            Reference time; true geocent time for GW is t_ref + theta["geocent_time"].
        time_marginalization_kwargs: dict
            Time marginalization parameters. If None, no time marginalization is used.
        calibration_marginalization_kwargs: dict
            Calibration marginalization parameters. If None, no calibration marginalization is used.
        phase_marginalization_kwargs: dict
            Phase marginalization parameters. If None, no phase marginalization is used.
        """
        super().__init__(
            wfg_kwargs=wfg_kwargs,
            wfg_domain=wfg_domain,
            data_domain=data_domain,
            ifo_list=list(event_data["waveform"].keys()),
            t_ref=t_ref,
        )

        self.asd = event_data["asds"]

        self.whitened_strains = {
            k: v / self.asd[k] / self.data_domain.noise_std
            for k, v in event_data["waveform"].items()
        }
        if len(list(self.whitened_strains.values())[0]) != data_domain.max_idx + 1:
            raise ValueError("Strain data does not match domain.")
        # log noise evidence, independent of theta and waveform model
        self.log_Zn = sum(
            [
                -1 / 2.0 * inner_product(d_ifo, d_ifo)
                for d_ifo in self.whitened_strains.values()
            ]
        )
        # For completeness (not used): there is a PSD-dependent contribution to the
        # likelihood,  which is typically ignored as it is constant for a given PSD
        # (e.g.,  for different waveform models) or event strains. Intuitively it is the
        # correction term that needs to be added to the log-likelihood to account
        # for the fact we compute N[0,1](strain/ASD) instead of N[0,ASD^2](strain).
        # But this contribution is typically ignored since we are only interested in
        # log-likelihood *differences*, see e.g. https://arxiv.org/pdf/1809.02293.pdf.
        # psi = - sum_i log(2pi * PSD_i) = - 2 * sum_i * log(2pi * ASD_i)
        # self.psi = -2 * sum(
        #     np.sum(np.log(2 * np.pi * asd)) for asd in self.asd.values()
        # )

        # Value in Veitch et al (2015)
        self.psi = -2 * np.sum(
            [
                np.sum(np.log(np.sqrt(2 * np.pi) * asd * self.data_domain.noise_std))
                for asd in self.asd.values()
            ]
        )
        self.whiten = True
        self.phase_grid = phase_grid

        # optionally initialize time marginalization
        self.time_marginalization = False
        if time_marginalization_kwargs is not None:
            self.initialize_time_marginalization(**time_marginalization_kwargs)

        # optionally initialize phase marginalization
        self.phase_marginalization = False
        if phase_marginalization_kwargs is not None:
            self.phase_marginalization = True
            # flag whether to use exp(2i * phi) approximation for phase transformations
            self.pm_approx_22_mode = phase_marginalization_kwargs.get(
                "approximation_22_mode", True
            )
            # if we don't use the exp(2i * phase) approximation, we need a phase grid
            if not self.pm_approx_22_mode:
                n_grid = phase_marginalization_kwargs.get("n_grid", 1_000)
                # use endpoint = False for grid, since phase = 0/2pi are equivalent
                self.phase_grid = np.linspace(0, 2 * np.pi, n_grid, endpoint=False)
            else:
                print("Using phase marginalization with (2,2) mode approximation.")

        # Initialize calibration marginalization using the setter from GWSignal.
        self.calibration_marginalization_kwargs = calibration_marginalization_kwargs

    def initialize_time_marginalization(self, t_lower, t_upper, n_fft=1):
        """
        Initialize time marginalization. Time marginalization can be performed via FFT,
        which is super fast. However, this limits the time resolution to delta_t =
        1/self.data_domain.f_max. In order to allow for a finer time resolution we
        compute the time marginalized likelihood n_fft via FFT on a grid of n_fft
        different time shifts [0, delta_t, 2*delta_t, ..., (n_fft-1)*delta_t] and
        average over the time shifts. The effective time resolution is thus

            delta_t_eff = delta_t / n_fft = 1 / (f_max * n_fft).

        Note: Time marginalization in only implemented for uniform time priors.

        Parameters
        ----------
        t_lower: float
            Lower time bound of the uniform time prior.
        t_upper: float
            Upper time bound of the uniform time prior.
        n_fft: int = 1
            Size of grid for FFT for time marginalization.
        """
        self.time_marginalization = True
        self.n_fft = n_fft
        delta_t = 1.0 / self.data_domain.f_max  # time resolution of FFT
        # time shifts for different FFTs
        self.t_FFT = np.arange(self.n_fft) * delta_t / self.n_fft

        self.shifted_strains = {}
        for idx, dt in enumerate(self.t_FFT):
            # Instead of shifting the waveform mu by + dt when computing the
            # time-marginalized likelihood, we shift the strain data by -dt. This saves
            # time for likelihood evaluations, since it can be precomputed.
            self.shifted_strains[dt] = {
                k: v * np.exp(-2j * np.pi * self.data_domain() * (-dt))
                for k, v in self.whitened_strains.items()
            }

        # Get the time prior. This will be multiplied with the result of the FFT.
        T = 1 / self.data_domain.delta_f
        time_axis = np.arange(len(self.data_domain())) / self.data_domain.f_max
        self.time_grid = time_axis[:, np.newaxis] + self.t_FFT[np.newaxis, :]
        active_indices = np.where(
            (self.time_grid >= t_lower) & (self.time_grid <= t_upper)
            | (self.time_grid - T >= t_lower) & (self.time_grid - T <= t_upper)
        )
        time_prior = np.zeros(self.time_grid.shape)
        time_prior[active_indices] = 1.0
        with np.errstate(divide="ignore"):  # ignore warnings for log(0) = -inf
            self.time_prior_log = np.log(time_prior / np.sum(time_prior))

    def log_likelihood(self, theta):
        if (
            sum(
                [
                    self.time_marginalization,
                    self.phase_marginalization,
                    bool(self.calibration_marginalization_kwargs),
                ]
            )
            > 1
        ):
            raise NotImplementedError(
                "Only one out of time-marginalization, phase-marginalization and calibration-marginalization "
                "can be used at a time."
            )

        if self.time_marginalization:
            return self._log_likelihood_time_marginalized(theta)
        elif self.phase_marginalization:
            return self._log_likelihood_phase_marginalized(theta)
        elif self.calibration_marginalization_kwargs:
            return self._log_likelihood_calibration_marginalized(theta)
        else:
            return self._log_likelihood(theta)

    def _log_likelihood(self, theta):
        """
        The likelihood is given by

                  log L(d|theta) = psi - 1/2. <d - mu(theta), d - mu(theta)>

        where psi is a waveform model independent (and thus irrelevant) constant. Here,
        we denote the strain data by d and the GW signal by mu(theta).
        [see e.g. arxiv.org/pdf/1809.02293, equation (44) for details]
        We expand this expression below to compute the log likelihood, and omit psi.
        The expaneded expression reads

                log L(d|theta) = log_Zn + kappa2(theta) - 1/2 rho2opt(theta),

                log_Zn = -1/2. <d, d>,
                kappa2(theta) = <d, mu(theta)>,
                rho2opt(theta) = <mu(theta), mu(theta)>.

        The noise-weighted inner product is defined as

                  <a, b> = 4 * delta_f * sum(a.conj() * b / PSD).real.

        Here, we work with data d and signals mu that are already whitened by
        1 / [sqrt(PSD) * domain.noise_std], where

                  noise_std = np.sqrt(window_factor) / np.sqrt(4 * delta_f).

        With this preprocessing, the inner products thus simply become

                  <a, b> = sum(a.conj() * b).real.

        ! Be careful with window factors here !


        Time marginalization:
        The above expansion of the likelihood is particularly useful for time
        marginalization, as only kappa2 depends on the time parameter.


        Parameters
        ----------
        theta: dict
            BBH parameters.

        Returns
        -------
        log_likelihood: float
        """

        # Step 1: Compute whitened GW strain mu(theta) for parameters theta.
        mu = self.signal(theta)["waveform"]
        d = self.whitened_strains

        # Step 2: Compute likelihood. log_Zn is precomputed, so we only need to
        # compute the remaining terms rho2opt and kappa2
        rho2opt = sum([inner_product(mu_ifo, mu_ifo) for mu_ifo in mu.values()])
        kappa2 = sum(
            [
                inner_product(d_ifo, mu_ifo)
                for d_ifo, mu_ifo in zip(d.values(), mu.values())
            ],
        )
        return self.log_Zn + kappa2 - 1 / 2.0 * rho2opt

    def log_likelihood_phase_grid(self, theta, phases=None):
        # TODO: Implement for time marginalization
        if self.phase_marginalization:
            raise ValueError(
                "Can't compute likelihood on a phase grid for "
                "phase-marginalized posteriors"
            )
        if self.time_marginalization:
            raise NotImplementedError(
                "log_likelihood on phase grid not yet implemented."
            )

        if self.waveform_generator.spin_conversion_phase != 0:
            raise ValueError(
                f"The log likelihood on a phase grid assumes "
                f"WaveformGenerator.spin_conversion_phase = 0, "
                f"got {self.waveform_generator.spin_conversion_phase}."
            )

        d = self.whitened_strains
        if phases is None:
            phases = self.phase_grid

        # Step 1: Compute signal for phase = 0, separated into the m-contributions from
        # the individual modes.
        pol_m = self.signal_m({**theta, "phase": 0})
        pol_m = {k: pol["waveform"] for k, pol in pol_m.items()}

        # Step 2: Precompute complex inner products (mu, mu) and (d, mu) for the
        # individual modes m.
        min_idx = self.data_domain.min_idx
        m_vals = sorted(pol_m.keys())

        # rho2opt is defined as the inner product of the waveform mu with itself.
        # Since we work with whitened data, the inner product simply reads
        #
        #   (mu, mu) = sum(mu.conj() * mu).real,
        #
        # where the sum extends over frequency bins and detectors. Applying phase
        # transformations exp(-i * m * phi) to the individual modes will lead to
        # cross terms (the ^-symbol indicates the phase shift by phi)
        #
        #   (mu^_m, mu^_n) = [(mu_m.conj() * mu_n) * exp(-i * (n - m) * phi)].real.
        #
        # Below we precompute the cross terms sum(mu_m.conj() * mu_n) and the constant
        # contribution for m = n.
        rho2opt_const = 0
        rho2opt_crossterms = {}
        for idx, m in enumerate(m_vals):
            mu_m = pol_m[m]
            # contribution to rho2opt_const
            rho2opt_const += sum(
                [inner_product(mu_ifo, mu_ifo, min_idx) for mu_ifo in mu_m.values()]
            )
            # cross terms
            for n in m_vals[idx + 1 :]:
                mu_n = pol_m[n]
                # factor 2, since (m, n) and (n, m) cross terms contribute symmetrically
                rho2opt_crossterms[(m, n)] = 2 * sum(
                    [
                        inner_product_complex(mu_m_ifo, mu_n_ifo, min_idx)
                        for mu_m_ifo, mu_n_ifo in zip(mu_m.values(), mu_n.values())
                    ]
                )
        # kappa2 is given by
        #
        #   (d, mu) = sum(d.conj() * mu).real.
        #
        # Applying phase transformations exp(-i * m * phi) to the individual modes
        # thus corresponds to
        #
        #   (d, mu^_m) = [(d.conj() * mu_m) * exp(-i * m * phi)].real.
        #
        # Below we precompute (d.conj() * mu_m) for the different modes m.
        kappa2_modes = {}
        for m in m_vals:
            mu_m = pol_m[m]
            kappa2_modes[m] = sum(
                [
                    inner_product_complex(d_ifo, mu_ifo, min_idx)
                    for d_ifo, mu_ifo in zip(d.values(), mu_m.values())
                ]
            )

        log_likelihoods = np.ones(len(phases))
        kappa2_all = []
        rho2opt_all = []
        for idx, phase in enumerate(phases):
            # get rho2opt
            rho2opt = rho2opt_const
            for (m, n), c in rho2opt_crossterms.items():
                rho2opt += (c * np.exp(-1j * (n - m) * phase)).real
            # get kappa2
            kappa2 = 0
            for m in m_vals:
                kappa2 += (kappa2_modes[m] * np.exp(-1j * m * phase)).real
            rho2opt_all.append(rho2opt)
            kappa2_all.append(kappa2)

            log_likelihoods[idx] = self.log_Zn + kappa2 - 1 / 2.0 * rho2opt

            # # comment out for cross check:
            # mu = sum_contributions_m(pol_m, phase_shift=phase)
            # rho2opt_ref = sum([inner_product(mu_ifo, mu_ifo) for mu_ifo in mu.values()])
            # kappa2_ref = sum(
            #     [
            #         inner_product(d_ifo, mu_ifo)
            #         for d_ifo, mu_ifo in zip(d.values(), mu.values())
            #     ]
            # )
            # assert rho2opt - rho2opt_ref < 1e-10
            # assert kappa2 - kappa2_ref < 1e-10

        # # Test that this works:
        # idx = len(phases) // 3
        # phase = phases[idx]
        # log_likelihood_ref = self.log_likelihood({**theta, "phase": phase})
        # print(log_likelihoods[idx] - log_likelihood_ref)

        return log_likelihoods

    def _log_likelihood_phase_marginalized(self, theta):
        """

        Parameters
        ----------
        theta: dict
            BBH parameters.

        Returns
        -------
        log_likelihood: float
        """

        # exp(2i * phase) approximation
        if self.pm_approx_22_mode:
            # Step 1: Compute whitened GW strain mu(theta) for parameters theta.
            # The phase parameter needs to be set to 0.
            theta["phase"] = 0.0
            mu = self.signal(theta)["waveform"]
            d = self.whitened_strains

            # Step 2: Compute likelihood. log_Zn is precomputed, so we only need to
            # compute the remaining terms rho2opt and kappa2
            rho2opt = sum([inner_product(mu_ifo, mu_ifo) for mu_ifo in mu.values()])
            # For the phase marginalized likelihood, we need to replace kappa2 with
            #
            #       log I0 ( |kappa2C| ),
            #
            # where kappa2C is the same as kappa2, but with the complex inner product
            # instead of the regular one, and I0 is the modified Bessel function of the
            # first kind of order 0.
            kappa2C = sum(
                [
                    inner_product_complex(d_ifo, mu_ifo)
                    for d_ifo, mu_ifo in zip(d.values(), mu.values())
                ]
            )
            return self.log_Zn + ln_i0(np.abs(kappa2C)) - 1 / 2.0 * rho2opt

        else:
            log_likelihoods_phase_grid = self.log_likelihood_phase_grid(theta)
            # Marginalize over phase. The phase-marginalized posterior is given by
            #
            #   p(theta^|d) = int d_phase p(d|theta^,phase) * p(theta^) * p(phase) / p(d),
            #
            #   where theta^ consists of all parameters except for the phase.
            #   To marginalize over the phase we evaluate it on a phase grid.
            #   With c = p(theta) / p(d) this yields
            #
            #   p(theta^|d) = c * int d_phase p(d|theta^,phase) * p(phase)
            #               = c * sum_i^N delta_phase_i p(theta^,i * delta_phase_i|d)
            #                   * p(i * delta_phase_i)
            #               = c * sum_i^N (2*pi/N) * p(theta^,i * delta_phase_i|d)
            #                   * (1/2*pi)
            #               = c * mean[p(theta^,i * delta_phase_i|d)].
            #
            # So to marginalize over the phase we simply need to take the mean of
            # np.exp(log_likelihoods_phase_grid). The constant c is accounted for
            # outside of this function.
            return logsumexp(log_likelihoods_phase_grid) - np.log(
                len(log_likelihoods_phase_grid)
            )

    def _log_likelihood_time_marginalized(self, theta):
        """
        Compute log likelihood with time marginalization.

        Parameters
        ----------
        theta

        Returns
        -------
        log_likelihood: float
        """
        # Step 1: Compute whitened GW strain mu(theta) for parameters theta.
        # The geocent_time parameter needs to be set to 0.
        theta["geocent_time"] = 0.0
        mu = self.signal(theta)["waveform"]
        # d = self.whitened_strains

        # Step 2: Compute likelihood. log_Zn is precomputed, so we only need to
        # compute the remaining terms rho2opt and kappa2.
        # rho2opt is time independent, and thus same as in the log_likelihood method.
        rho2opt = sum([inner_product(mu_ifo, mu_ifo) for mu_ifo in mu.values()])

        # kappa2 is time dependent. We use FFT to compute it for the discretized times
        # k * (delta_t/n_fft) and then sum over the time bins. The kappa2 contribution
        # is then given by
        #
        #       log sum_k exp(kappa2_k + log_prior_k),
        #
        # see Eq. (52) in https://arxiv.org/pdf/1809.02293.pdf. Here, kappa2_k is the
        # value of kappa2 and log_prior_k is the log_prior density at time
        # k * (delta_t/n_fft). The sum over k is the discretized integration of t.
        # Note: the time is discretized in two ways; for each FFT j (n_fft in total),
        # there are len(data_domain) time samples i, such that
        #
        #       t_ij = i * delta_t + j * (delta_t/n_fft).
        #
        # Summing over the time bins corresponds to a sum across both axes i and j.
        kappa2_ij = np.zeros((len(self.data_domain), self.n_fft))
        for j, dt in enumerate(self.t_FFT):
            # Get precomputed whitened strain, that is shifted by -dt.
            d = self.shifted_strains[dt]
            # Compute kappa2 contribution
            kappa2_ = [
                fft(d_ifo.conj() * mu_ifo).real
                for d_ifo, mu_ifo in zip(d.values(), mu.values())
            ]
            # sum contributions of different ifos
            kappa2_ij[:, j] = np.sum(kappa2_, axis=0)
        # Marginalize over time; this requires multiplying the likelihoods with the
        # prior (*not* in log space), summing over the time bins (both axes i and j!),
        # and then taking the log. See Eq. (52) in https://arxiv.org/pdf/1809.02293.pdf.
        # To prevent numerical issues, we use the logsumexp trick.
        assert kappa2_ij.shape == self.time_prior_log.shape
        exponent = kappa2_ij + self.time_prior_log
        alpha = np.max(exponent)
        kappa2 = alpha + np.log(np.sum(np.exp(exponent - alpha)))

        return self.log_Zn + kappa2 - 1 / 2.0 * rho2opt

    def _log_likelihood_calibration_marginalized(self, theta):
        """
        Computes log likelihood with calibration_marginalization

        Parameters
        ----------
        theta: dict
            BBH parameters.

        Returns
        -------
        log_likelihood: float
        """

        # Step 1: Compute whitened GW strain mu(theta) for parameters theta.
        mu = self.signal(theta)["waveform"]
        d = self.whitened_strains

        # In the calibration-marginalization case, the values of mu for each detector
        # have an additional leading dimension, which corresponds to the calibration
        # draws. These are averaged over in computing the likelihood.
        # TODO: Combine with _log_likelihood() into a single method.

        # Step 2: Compute likelihood. log_Zn is precomputed, so we only need to
        # compute the remaining terms rho2opt and kappa2

        rho2opt = np.sum(
            [inner_product(mu_ifo.T, mu_ifo.T) for mu_ifo in mu.values()], axis=0
        )
        kappa2 = np.sum(
            [
                inner_product(d_ifo[:, None], mu_ifo.T)
                for d_ifo, mu_ifo in zip(d.values(), mu.values())
            ],
            axis=0,
        )

        likelihoods = self.log_Zn + kappa2 - 1 / 2.0 * rho2opt
        # Return the average over calibration envelopes
        return logsumexp(likelihoods) - np.log(len(likelihoods))

    def d_inner_h_complex_multi(
        self, theta: pd.DataFrame, num_processes: int = 1
    ) -> np.ndarray:
        """
        Calculate the complex inner product (d | h(theta)) between the stored data d
        and a simulated waveform with given parameters theta. Works with multiprocessing.

        Parameters
        ----------
        theta : pd.DataFrame
            Parameters at which to evaluate h.
        num_processes : int
            Number of parallel processes to use.

        Returns
        -------
        complex : Inner product
        """
        with threadpool_limits(limits=1, user_api="blas"):
            # Generator object for theta rows. For idx this yields row idx of
            # theta dataframe, converted to dict, ready to be passed to
            # self.log_likelihood.
            theta_generator = (d[1].to_dict() for d in theta.iterrows())

            if num_processes > 1:
                with Pool(processes=num_processes) as pool:
                    d_inner_h_complex = pool.map(
                        self.d_inner_h_complex, theta_generator
                    )
            else:
                d_inner_h_complex = list(map(self.d_inner_h_complex, theta_generator))

        return np.array(d_inner_h_complex)

    def d_inner_h_complex(self, theta):
        """
        Calculate the complex inner product (d | h(theta)) between the stored data d
        and a simulated waveform with given parameters theta.

        Parameters
        ----------
        theta : dict
            Parameters at which to evaluate h.

        Returns
        -------
        complex : Inner product
        """
        # TODO: Implement for time marginalization.
        return self._d_inner_h_complex(theta)

    def _d_inner_h_complex(self, theta):
        mu = self.signal(theta)["waveform"]
        d = self.whitened_strains
        return sum(
            [
                inner_product_complex(d_ifo, mu_ifo)
                for d_ifo, mu_ifo in zip(d.values(), mu.values())
            ]
        )


def inner_product(a, b, min_idx=0, delta_f=None, psd=None):
    """
    Compute the inner product between two complex arrays. There are two modes: either,
    the data a and b are not whitened, in which case delta_f and the psd must be
    provided. Alternatively, if delta_f and psd are not provided, the data a and b are
    assumed to be whitened already (i.e., whitened as d -> d * sqrt(4 delta_f / psd)).

    Note: sum is only taken along axis 0 (which is assumed to be the frequency axis),
    while other axes are preserved. This is e.g. useful when evaluating kappa2 on a
    phase grid.

    Parameters
    ----------
    a: np.ndaarray
        First array with frequency domain data.
    b: np.ndaarray
        Second array with frequency domain data.
    min_idx: int = 0
        Truncation of likelihood integral, index of lowest frequency bin to consider.
    delta_f: float
        Frequency resolution of the data. If None, a and b are assumed to be whitened
        and the inner product is computed without further whitening.
    psd: np.ndarray = None
        PSD of the data. If None, a and b are assumed to be whitened and the inner
        product is computed without further whitening.

    Returns
    -------
    inner_product: float
    """
    #
    if psd is not None:
        if delta_f is None:
            raise ValueError(
                "If unwhitened data is provided, both delta_f and psd must be provided."
            )
        return 4 * delta_f * np.sum((a.conj() * b / psd)[min_idx:], axis=0).real
    else:
        return np.sum((a.conj() * b)[min_idx:], axis=0).real


def inner_product_complex(a, b, min_idx=0, delta_f=None, psd=None):
    """
    Same as inner product, but without taking the real part. Retaining phase
    information is useful for the phase-marginalized likelihood. For further
    documentation see inner_product function.
    """
    #
    if psd is not None:
        if delta_f is None:
            raise ValueError(
                "If unwhitened data is provided, both delta_f and psd must be provided."
            )
        return 4 * delta_f * np.sum((a.conj() * b / psd)[min_idx:], axis=0)
    else:
        return np.sum((a.conj() * b)[min_idx:], axis=0)


def build_stationary_gaussian_likelihood(
    metadata,
    event_dataset=None,
    time_marginalization_kwargs=None,
):
    """
    Build a StationaryGaussianLikelihoodBBH object from the metadata.

    Parameters
    ----------
    metadata: dict
        Metadata from stored dingo parameter samples file.
        Typially accessed via pd.read_pickle(/path/to/dingo-output.pkl).metadata.
    event_dataset: str = None
        Path to event dataset for caching. If None, don't cache.
    time_marginalization_kwargs: dict = None
        Forwarded to the likelihood.

    Returns
    -------
    likelihood: StationaryGaussianGWLikelihood
        likelihood object
    """
    # get strain data
    event_data, data_domain = get_event_data_and_domain(
        metadata["model"], event_dataset=event_dataset, **metadata["event"]
    )

    # set up likelihood
    likelihood = StationaryGaussianGWLikelihood(
        wfg_kwargs=metadata["model"]["dataset_settings"]["waveform_generator"],
        wfg_domain=build_domain(metadata["model"]["dataset_settings"]["domain"]),
        data_domain=data_domain,
        event_data=event_data,
        t_ref=metadata["event"]["time_event"],
        time_marginalization_kwargs=time_marginalization_kwargs,
    )

    return likelihood


def get_wfg(wfg_kwargs, data_domain, frequency_range=None):
    """
    Set up waveform generator from wfg_kwargs. The domain of the wfg is primarily
    determined by the data domain, but a new (larger) frequency range can be
    specified if this is necessary for the waveforms to be generated successfully
    (e.g., for EOB waveforms which require a sufficiently small f_min and sufficiently
    large f_max).

    Parameters
    ----------
    wfg_kwargs: dict
        Waveform generator parameters.
    data_domain: dingo.gw.domains.Domain
        Domain of event data, with bounds determined by likelihood integral.
    frequency_range: dict = None
        Frequency range for waveform generator. If None, that of data domain is used,
        which corresponds to the bounds of the likelihood integral.
        Possible keys:
            'f_start': float
                Frequency at which to start the waveform generation. Overrides f_start in
                metadata["model"]["dataset_settings"]["waveform_generator"].
            'f_end': float
                Frequency at which to start the waveform generation.

    Returns
    -------
    wfg: dingo.gw.waveform_generator.WaveformGenerator
        Waveform generator object.

    """
    if frequency_range is None:
        return WaveformGenerator(domain=data_domain, **wfg_kwargs)

    else:
        if "f_start" in frequency_range and frequency_range["f_start"] is not None:
            if frequency_range["f_start"] > data_domain.f_min:
                raise ValueError("f_start must be less than f_min.")
            wfg_kwargs["f_start"] = frequency_range["f_start"]
        if "f_end" in frequency_range and frequency_range["f_end"] is not None:
            if frequency_range["f_end"] < data_domain.f_max:
                raise ValueError("f_end must be greater than f_max.")
            # get wfg domain, but care to not modify the original data_domain
            data_domain = build_domain(
                {**data_domain.domain_dict, "f_max": frequency_range["f_end"]}
            )
        return WaveformGenerator(domain=data_domain, **wfg_kwargs)


def main():
    import pandas as pd

    samples = pd.read_pickle(
        "/Users/maxdax/Documents/Projects/GW-Inference/dingo/datasets/dingo_samples"
        "/02_XPHM/dingo_samples_GW150914.pkl"
    )
    event_dataset = (
        "/Users/maxdax/Documents/Projects/GW-Inference/dingo/dingo-devel"
        "/tutorials/02_gwpe/datasets/strain_data/events_dataset.hdf5"
    )

    likelihood = build_stationary_gaussian_likelihood(samples.attrs, event_dataset)

    from tqdm import tqdm

    log_likelihoods = []
    for idx in tqdm(range(1000)):
        theta = dict(samples.iloc[idx])
        try:
            l = likelihood.log_prob(theta)
        except:
            print(idx)
            l = float("nan")
        log_likelihoods.append(l)
    log_likelihoods = np.array(log_likelihoods)
    log_likelihoods = log_likelihoods[~np.isnan(log_likelihoods)]
    print(f"mean: {np.mean(log_likelihoods)}")
    print(f"std: {np.std(log_likelihoods)}")
    print(f"max: {np.max(log_likelihoods)}")
    print(f"min: {np.min(log_likelihoods)}")


if __name__ == "__main__":
    main()



import pytest
import numpy as np

from dingo.gw.transforms import UnpackDict

def test_UnpackDict():
    sample = {'a': 10, 'b': np.random.rand(100), 'c': None}
    unpack_dict = UnpackDict(['b', 'a'])
    b, a = unpack_dict(sample)
    assert id(b) == id(sample['b'])
    assert a == sample['a']


import math
import numpy as np
import torch
import pandas as pd
from bilby.gw.detector.interferometer import Interferometer
from lal import GreenwichMeanSiderealTime
from typing import Union
from bilby.gw.detector import calibration
from bilby.gw.prior import CalibrationPriorDict


CC = 299792458.0


def time_delay_from_geocenter(
    ifo: Interferometer,
    ra: Union[float, np.ndarray, torch.Tensor],
    dec: Union[float, np.ndarray, torch.Tensor],
    time: float,
):
    """
    Calculate time delay between ifo and geocenter. Identical to method
    ifo.time_delay_from_geocenter(ra, dec, time), but the present implementation allows
    for batched computation, i.e., it also accepts arrays and tensors for ra and dec.

    Implementation analogous to bilby-cython implementation
    https://git.ligo.org/colm.talbot/bilby-cython/-/blob/main/bilby_cython/geometry.pyx,
    which is in turn based on XLALArrivaTimeDiff in TimeDelay.c.

    Parameters
    ----------
    ifo: bilby.gw.detector.interferometer.Interferometer
        bilby interferometer object.
    ra: Union[float, np.array, torch.Tensor]
        Right ascension of the source in radians. Either float, or float array/tensor.
    dec: Union[float, np.array, torch.Tensor]
        Declination of the source in radians. Either float, or float array/tensor.
    time: float
        GPS time in the geocentric frame.

    Returns
    -------
    float: Time delay between the two detectors in the geocentric frame
    """
    # check that ra and dec are of same type and length
    if type(ra) != type(dec):
        raise ValueError(
            f"ra type ({type(ra)}) and dec type ({type(dec)}) don't match."
        )
    if isinstance(ra, (np.ndarray, torch.Tensor)):
        if len(ra.shape) != 1:
            raise ValueError(f"Only one axis expected for ra and dec, got multiple.")
        if ra.shape != dec.shape:
            raise ValueError(
                f"Shapes of ra ({ra.shape}) and dec ({dec.shape}) don't match."
            )

    if isinstance(ra, (float, np.float32, np.float64)):
        return ifo.time_delay_from_geocenter(ra, dec, time)

    elif isinstance(ra, (np.ndarray, torch.Tensor)) and len(ra) == 1:
        return ifo.time_delay_from_geocenter(ra[0], dec[0], time)

    else:
        if isinstance(ra, np.ndarray):
            sin = np.sin
            cos = np.cos
        elif isinstance(ra, torch.Tensor):
            sin = torch.sin
            cos = torch.cos
        else:
            raise NotImplementedError(
                "ra, dec must be either float, np.ndarray, or torch.Tensor."
            )

        gmst = math.fmod(GreenwichMeanSiderealTime(float(time)), 2 * np.pi)
        phi = ra - gmst
        theta = np.pi / 2 - dec
        sintheta = sin(theta)
        costheta = cos(theta)
        sinphi = sin(phi)
        cosphi = cos(phi)
        detector_1 = ifo.vertex
        detector_2 = np.zeros(3)
        return (
            (detector_2[0] - detector_1[0]) * sintheta * cosphi
            + (detector_2[1] - detector_1[1]) * sintheta * sinphi
            + (detector_2[2] - detector_1[2]) * costheta
        ) / CC


class GetDetectorTimes(object):
    """
    Compute the time shifts in the individual detectors based on the sky
    position (ra, dec), the geocent_time and the ref_time.
    """

    def __init__(self, ifo_list, ref_time):
        self.ifo_list = ifo_list
        self.ref_time = ref_time

    def __call__(self, input_sample):
        sample = input_sample.copy()
        # the line below is required as sample is a shallow copy of
        # input_sample, and we don't want to modify input_sample
        extrinsic_parameters = sample["extrinsic_parameters"].copy()
        ra = extrinsic_parameters["ra"]
        dec = extrinsic_parameters["dec"]
        geocent_time = extrinsic_parameters["geocent_time"]
        for ifo in self.ifo_list:
            if type(ra) == torch.Tensor:
                # computation does not work on gpu, so do it on cpu
                ra = ra.cpu()
                dec = dec.cpu()
            dt = time_delay_from_geocenter(ifo, ra, dec, self.ref_time)
            if type(dt) == torch.Tensor:
                dt = dt.to(geocent_time.device)
            ifo_time = geocent_time + dt
            extrinsic_parameters[f"{ifo.name}_time"] = ifo_time
        sample["extrinsic_parameters"] = extrinsic_parameters
        return sample


class ProjectOntoDetectors(object):
    """
    Project the GW polarizations onto the detectors in ifo_list. This does
    not sample any new parameters, but relies on the parameters provided in
    sample['extrinsic_parameters']. Specifically, this transform applies the
    following operations:

    (1) Rescale polarizations to account for sampled luminosity distance
    (2) Project polarizations onto the antenna patterns using the ref_time and
        the extrinsic parameters (ra, dec, psi)
    (3) Time shift the strains in the individual detectors according to the
        times <ifo.name>_time provided in the extrinsic parameters.
    """

    def __init__(self, ifo_list, domain, ref_time):
        self.ifo_list = ifo_list
        self.domain = domain
        self.ref_time = ref_time

    def __call__(self, input_sample):
        sample = input_sample.copy()
        # the line below is required as sample is a shallow copy of
        # input_sample, and we don't want to modify input_sample
        parameters = sample["parameters"].copy()
        extrinsic_parameters = sample["extrinsic_parameters"].copy()
        try:
            d_ref = parameters["luminosity_distance"]
            d_new = extrinsic_parameters.pop("luminosity_distance")
            ra = extrinsic_parameters.pop("ra")
            dec = extrinsic_parameters.pop("dec")
            psi = extrinsic_parameters.pop("psi")
            tc_ref = parameters["geocent_time"]
            assert tc_ref == 0, (
                "This should always be 0. If for some reason "
                "you want to save time shifted polarizations,"
                " then remove this assert statement."
            )
            tc_new = extrinsic_parameters.pop("geocent_time")
        except:
            raise ValueError("Missing parameters.")

        # (1) rescale polarizations and set distance parameter to sampled value
        hc = sample["waveform"]["h_cross"] * d_ref / d_new
        hp = sample["waveform"]["h_plus"] * d_ref / d_new
        parameters["luminosity_distance"] = d_new

        strains = {}
        for ifo in self.ifo_list:
            # (2) project strains onto the different detectors
            fp = ifo.antenna_response(ra, dec, self.ref_time, psi, mode="plus")
            fc = ifo.antenna_response(ra, dec, self.ref_time, psi, mode="cross")
            strain = fp * hp + fc * hc

            # (3) time shift the strain. If polarizations are timeshifted by
            #     tc_ref != 0, undo this here by subtracting it from dt.
            dt = extrinsic_parameters[f"{ifo.name}_time"] - tc_ref
            strains[ifo.name] = self.domain.time_translate_data(strain, dt)

        # Add extrinsic parameters corresponding to the transformations
        # applied in the loop above to parameters. These have all been popped off of
        # extrinsic_parameters, so they only live one place.
        parameters["ra"] = ra
        parameters["dec"] = dec
        parameters["psi"] = psi
        parameters["geocent_time"] = tc_new
        for ifo in self.ifo_list:
            param_name = f"{ifo.name}_time"
            parameters[param_name] = extrinsic_parameters.pop(param_name)

        sample["waveform"] = strains
        sample["parameters"] = parameters
        sample["extrinsic_parameters"] = extrinsic_parameters

        return sample


class TimeShiftStrain(object):
    """
    Time shift the strains in the individual detectors according to the
    times <ifo.name>_time provided in the extrinsic parameters.
    """

    def __init__(self, ifo_list, domain):
        self.ifo_list = ifo_list
        self.domain = domain

    def __call__(self, input_sample):
        sample = input_sample.copy()
        extrinsic_parameters = input_sample["extrinsic_parameters"].copy()

        strains = {}

        if isinstance(input_sample["waveform"], dict):
            for ifo in self.ifo_list:
                # time shift the strain
                strain = input_sample["waveform"][ifo.name]
                dt = extrinsic_parameters.pop(f"{ifo.name}_time")
                strains[ifo.name] = self.domain.time_translate_data(strain, dt)

        elif isinstance(input_sample["waveform"], torch.Tensor):
            strains = input_sample["waveform"]
            dt = [extrinsic_parameters.pop(f"{ifo.name}_time") for ifo in self.ifo_list]
            dt = torch.stack(dt, 1)
            strains = self.domain.time_translate_data(strains, dt)

        else:
            raise NotImplementedError(
                f"Unexpected type {type(input_sample['waveform'])}, expected dict or "
                f"torch.Tensor"
            )

        sample["waveform"] = strains
        sample["extrinsic_parameters"] = extrinsic_parameters

        return sample


class ApplyCalibrationUncertainty(object):
    r"""
    Expand out a waveform using several detector calibration draws. These multiple
    draws are intended to be used for marginalizing over calibration uncertainty.

    Detector calibration uncertainty is modeled as described in
    https://dcc.ligo.org/LIGO-T1400682/public

    Gravitational wave data $d$ is assumed to be of the form

    $$d(f) = h_{obs}(f) + n(f),$$

    where $h_{obs}$ is the observed waveform and $n$ is the noise. Since the detector
    is not perfectly calibrated, the observed waveform is not identical to the true
    waveform $h(f)$. Rather, it is assumed to have corrections of the form

    $$h_{obs}(f) = h(f) * (1 + \delta A(f)) * \exp(i \delta \phi(f)),$$

    where $\delta A(f)$ and $\delta \phi(f)$ are frequency-dependent amplitude and
    phase errors. Under the calibration model, these are parametrized with cubic
    splines, defined in terms of calibration parameters $A_i$ and $\phi_i$, defined
    at log-spaced frequency nodes,

    $$
    \delta A(f) &= \mathrm{spline}(f; {f_i, \delta A_i}), \\
    \delta \phi(f) &= \mathrm{spline}(f; {f_i, \delta \phi_i}).
    $$

    The calibration parameters are not known precisely, rather they are assumed to be
    normally distributed, with mean 0 and standard deviation  determined by the
    "calibration envelope", which varies from event to event.

    For each detector waveform, this transform draws a collection of $N$
    calibration curves $\{(\delta A^n(f), \delta \phi^n(f))\}_{n=1}^N$ according to a
    calibration envelope, and applies them to generate $N$ observed waveforms $\{h^n_{
    obs}(f)\}$. This is intended to be used for marginalizing over the calibration
    uncertainty when evaluating the likelihood for importance sampling.
    """

    def __init__(
        self,
        ifo_list,
        data_domain,
        calibration_envelope,
        num_calibration_curves,
        num_calibration_nodes,
    ):
        r"""
        Parameters
        ---------

        ifo_list : InterferometerList
            List of Interferometers present in the analysis.
        data_domain : Domain
            Domain on which data is defined.
        calibration_envelope : dict
            Dictionary of the form ``{"H1": filepath, "L1": filepath}``,
            where the filepaths are strings pointing to ".txt" files containing
            calibration envelopes. The calibration envelope depends on the event analyzed,
            and therefore  remains fixed for all applications of the transform. The
            calibration envelope is used to define the variances $(\sigma_{\delta A_i},
            \sigma_{\delta \phi_i})$ of the calibration paramters.
        num_calibration_curves : int
            Number of calibration curves $N$ to produce and apply to the
            waveform. Ultimately, this will translate to the number of samples in the
            Monte Carlo estimate of the marginalized likelihood integral.
        num_calibration_nodes : int
            Number of log-spaced frequency nodes $f_i$ to use in defining the spline.
        """

        self.ifo_list = ifo_list
        self.num_calibration_curves = num_calibration_curves

        self.data_domain = data_domain
        self.calibration_prior = {}
        if all([s.endswith(".txt") for s in calibration_envelope.values()]):
            # Generating .h5 lookup table from priors in .txt file
            self.calibration_envelope = calibration_envelope
            for ifo in self.ifo_list:
                # Setting calibration model to cubic spline
                ifo.calibration_model = calibration.CubicSpline(
                    f"recalib_{ifo.name}_",
                    minimum_frequency=data_domain.f_min,
                    maximum_frequency=data_domain.f_max,
                    n_points=num_calibration_nodes,
                )

                # Setting priors
                # What this will do is take the the calibration envelope and set
                # a spline on the median and sigma of the amplitude and phase.
                # Then in log frequency it will setup node points say at
                # frequency points, $f_i$.  Then for each node point f_i, it
                # will create a gaussian prior according to the spline of the
                # median and sigma found earlier
                self.calibration_prior[
                    ifo.name
                ] = CalibrationPriorDict.from_envelope_file(
                    self.calibration_envelope[ifo.name],
                    self.data_domain.f_min,
                    self.data_domain.f_max,
                    num_calibration_nodes,
                    ifo.name,
                )

        else:
            raise Exception("Calibration envelope must be specified in a .txt file!")

    def __call__(self, input_sample):
        sample = input_sample.copy()
        for ifo in self.ifo_list:
            calibration_parameter_draws, calibration_draws = {}, {}
            # Sampling from prior
            calibration_parameter_draws[ifo.name] = pd.DataFrame(
                self.calibration_prior[ifo.name].sample(self.num_calibration_curves)
            )
            calibration_draws[ifo.name] = np.zeros(
                (
                    self.num_calibration_curves,
                    len(self.data_domain.sample_frequencies),
                ),
                dtype=complex,
            )

            for i in range(self.num_calibration_curves):
                calibration_draws[ifo.name][
                    i, self.data_domain.frequency_mask
                ] = ifo.calibration_model.get_calibration_factor(
                    self.data_domain.sample_frequencies[
                        self.data_domain.frequency_mask
                    ],
                    prefix="recalib_{}_".format(ifo.name),
                    **calibration_parameter_draws[ifo.name].iloc[i],
                )

            # Multiplying the sample waveform in the interferometer according to
            # the calibration curve.  This is done by following the perscription
            # here:
            #
            # https://dcc.ligo.org/LIGO-T1400682 Eq 3 and 4
            #
            # We take the waveform h(f) and multiply it by C = (1 + \delta A(f))
            # \exp(i \delta \psi) i.e. h_obs(f) = C * h(f)
            # Here C is "calibration_draws"

            # Padding 0's to everything in the calibration array which is below f_min

            sample["waveform"][ifo.name] = (
                sample["waveform"][ifo.name] * calibration_draws[ifo.name]
            )

        return sample



import os
import argparse
from typing import Dict
from pycondor import Job, Dagman
import yaml
import copy

# Fixed file names
svd_fn = "svd.hdf5"
settings_part_fn = "settings_part.yaml"
settings_svd_part_fn = "settings_svd_part.yaml"
svd_dataset_part_prefix = "svd_dataset_part_"
svd_dataset_fn = "svd_dataset.hdf5"
dataset_part_prefix = "dataset_part_"


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--settings_file",
        type=str,
        required=True,
        help="YAML file containing database settings",
    )
    parser.add_argument(
        "--out_file",
        type=str,
        default="waveform_dataset.hdf5",
        help="Name of file for storing dataset.",
    )
    parser.add_argument(
        "--num_jobs",
        type=int,
        required=True,
        help="Number of condor jobs over which to split work.",
    )
    parser.add_argument(
        "--temp_dir",
        type=str,
        default="temp",
        help="Directory for storing temporary files.",
    )
    parser.add_argument(
        "--env_path",
        type=str,
        required=True,
        help="Absolute path to the dingo Python environment. "
        'We will execute scripts in "env_path/bin/".',
    )

    # condor arguments
    parser.add_argument("--request_cpus", type=int, default=None, help="CPUs per job.")
    parser.add_argument(
        "--request_memory", type=int, default=None, help="Memory per job."
    )
    parser.add_argument(
        "--request_memory_high", type=int, default=None, help="Memory per job, "
                                                              "used only for "
                                                              "aggregating data and "
                                                              "building SVD."
    )
    parser.add_argument("--error", type=str, default="condor/error")
    parser.add_argument("--output", type=str, default="condor/output")
    parser.add_argument("--log", type=str, default="condor/log")
    parser.add_argument("--submit", type=str, default="condor/submit")

    return parser.parse_args()


def modulus_check(a: int, b: int, a_label: str, b_label: str):
    """Raise error if a % b != 0."""
    if a % b != 0:
        raise ValueError(
            f"Expected {a_label} mod {b_label} to be zero. "
            f"But got {a} mod {b} = {a % b}."
        )


def create_args_string(args_dict: Dict):
    """Generate argument string from dictionary of argument names and arguments."""
    return "".join([f"--{k} {v} " for k, v in args_dict.items()])


def configure_runs(settings, num_jobs, temp_dir):
    """Prepare and save settings .yaml files for generating subsets of the dataset.
    Generally this will produce two .yaml files, one for generating the main dataset,
    one for the SVD training.

    Parameters
    ----------
    settings : dict
        Settings for full dataset configuration.
    num_jobs : int
        Number of jobs over which to split the run.
    temp_dir : str
        Name of (temporary) directory in which to place temporary output files.
    """

    settings_part = copy.deepcopy(settings)
    num_samples = settings["num_samples"]
    modulus_check(num_samples, num_jobs, "num_samples", "num_jobs")
    settings_part["num_samples"] = num_samples // num_jobs

    if "compression" in settings:
        if "svd" in settings["compression"]:

            # Configure a set of runs to generate waveforms from which to construct the
            # SVD. These should be uncompressed waveforms.
            settings_svd_part = copy.deepcopy(settings)
            num_samples = settings["compression"]["svd"]["num_training_samples"]
            num_samples += settings["compression"]["svd"].get(
                "num_validation_samples", 0
            )
            modulus_check(
                num_samples,
                num_jobs,
                "(number of SVD training + validation samples)",
                "num_jobs",
            )
            settings_svd_part["num_samples"] = num_samples // num_jobs
            del settings_svd_part["compression"]["svd"]
            with open(os.path.join(temp_dir, settings_svd_part_fn), "w") as f:
                yaml.dump(
                    settings_svd_part, f, default_flow_style=False, sort_keys=False
                )

            # Set the runs for the main dataset to use the saved SVD basis (assume it was
            # produced already, based on the above).
            settings_part["compression"]["svd"] = {
                "file": os.path.join(temp_dir, svd_fn)
            }

    with open(os.path.join(temp_dir, settings_part_fn), "w") as f:
        yaml.dump(settings_part, f, default_flow_style=False, sort_keys=False)


def create_dag(args, settings):
    """
    Create a Condor DAG from command line arguments to carry out the five steps in the
    workflow.

    """
    kwargs = {
        "request_cpus": args.request_cpus,
        "request_memory": args.request_memory,
        "submit": args.submit,
        "error": args.error,
        "output": args.output,
        "log": args.log,
        "getenv": True,
    }
    kwargs_high_memory = kwargs.copy()
    if args.request_memory_high is not None:
        kwargs_high_memory["request_memory"] = args.request_memory_high

    # scripts are installed in the env's bin directory
    path = os.path.join(args.env_path, "bin")
    temp_dir = args.temp_dir

    # DAG ---------------------------------------------------------------------
    dagman = Dagman(name="dingo_generate_dataset_dagman", submit=args.submit)

    # 1. Prepare SVD basis ----------------------------------------------------
    # This is only needed if we are using SVD compression
    if "compression" in settings:
        if "svd" in settings["compression"]:

            # --- (a) Generate dataset for SVD training. Split this over multiple jobs.
            executable = os.path.join(path, "dingo_generate_dataset")
            args_dict = {
                "settings_file": os.path.join(temp_dir, settings_svd_part_fn),
                "num_processes": args.request_cpus,
                "out_file": os.path.join(
                    temp_dir, svd_dataset_part_prefix + "$(Process).hdf5"
                ),
            }
            args_str = create_args_string(args_dict)
            generate_svd_dataset_part = Job(
                name="generate_svd_dataset_part",
                executable=executable,
                queue=args.num_jobs,
                dag=dagman,
                arguments=args_str,
                **kwargs,
            )

            # --- (b) Consolidate dataset.
            executable = os.path.join(path, "dingo_merge_datasets")
            args_dict = {
                "prefix": os.path.join(temp_dir, svd_dataset_part_prefix),
                "num_parts": args.num_jobs,
                "out_file": os.path.join(temp_dir, svd_dataset_fn),
            }
            args_str = create_args_string(args_dict)
            consolidate_svd_dataset = Job(
                name="consolidate_svd_dataset",
                executable=executable,
                dag=dagman,
                arguments=args_str,
                **kwargs_high_memory,
            )
            consolidate_svd_dataset.add_parent(generate_svd_dataset_part)

            # --- (c) Build SVD basis
            executable = os.path.join(path, "dingo_build_svd")
            args_dict = {
                "dataset_file": os.path.join(temp_dir, svd_dataset_fn),
                "size": settings["compression"]["svd"]["size"],
                "out_file": os.path.join(temp_dir, svd_fn),
                "num_train": settings["compression"]["svd"]["num_training_samples"],
            }
            args_str = create_args_string(args_dict)
            build_svd = Job(
                name="build_svd",
                executable=executable,
                dag=dagman,
                arguments=args_str,
                **kwargs_high_memory,
            )
            build_svd.add_parent(consolidate_svd_dataset)

    # 2. Prepare main dataset -------------------------------------------------

    # --- (a) Generate dataset. Split over multiple jobs.
    executable = os.path.join(path, "dingo_generate_dataset")
    args_dict = {
        "settings_file": os.path.join(temp_dir, settings_part_fn),
        "num_processes": args.request_cpus,
        "out_file": os.path.join(temp_dir, dataset_part_prefix + "$(Process).hdf5"),
    }
    args_str = create_args_string(args_dict)
    generate_dataset_part = Job(
        name="generate_dataset_part",
        executable=executable,
        queue=args.num_jobs,
        dag=dagman,
        arguments=args_str,
        **kwargs,
    )
    if "compression" in settings:
        if "svd" in settings["compression"]:
            generate_dataset_part.add_parent(build_svd)

    # --- (b) Consolidate dataset
    executable = os.path.join(path, "dingo_merge_datasets")
    args_dict = {
        "prefix": os.path.join(temp_dir, dataset_part_prefix),
        "num_parts": args.num_jobs,
        "out_file": args.out_file,
        "settings_file": args.settings_file,
    }
    args_str = create_args_string(args_dict)
    consolidate_dataset = Job(
        name="consolidate_dataset",
        executable=executable,
        dag=dagman,
        arguments=args_str,
        **kwargs_high_memory,
    )
    consolidate_dataset.add_parent(generate_dataset_part)

    return dagman


def main():
    args = parse_args()

    # Load settings
    with open(args.settings_file, "r") as f:
        settings = yaml.safe_load(f)

    # create temporary directory
    if not os.path.exists(args.temp_dir):
        os.mkdir(args.temp_dir)

    # Set up component .yaml files
    configure_runs(settings, args.num_jobs, args.temp_dir)

    dagman = create_dag(args, settings)

    try:
        dagman.visualize("waveform_dataset_generation_workflow.png")
    except:
        pass

    dagman.build()
    print(f"DAG submission file written to {args.submit}.")


if __name__ == "__main__":
    main()



import torch

from dingo.gw.domains import FrequencyDomain, TimeDomain, build_domain
from dingo.gw.gwutils import get_window_factor
import pytest
import numpy as np


@pytest.fixture
def uniform_FD_params():
    f_min = 20.0
    f_max = 4096.0
    delta_f = 1.0 / 4.0
    return {"f_min": f_min, "f_max": f_max, "delta_f": delta_f}


@pytest.fixture
def window_setup():
    type = "tukey"
    f_s = 4096
    T = 8.0
    roll_off = 0.4
    window_kwargs = {
        "type": type,
        "f_s": f_s,
        "T": T,
        "roll_off": roll_off,
    }
    window_factor = get_window_factor(window_kwargs)
    return window_kwargs, window_factor


def test_uniform_FD(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    n = int(p["f_max"] / p["delta_f"]) + 1
    frequencies_expected = np.linspace(0, p["f_max"], n)
    frequencies = domain()
    assert np.linalg.norm(frequencies - frequencies_expected) < 1e-15


def test_uniform_FD_mask(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    mask = domain.frequency_mask
    n_masked = int((p["f_max"] - p["f_min"]) / p["delta_f"]) + 1
    frequencies_expected_masked = np.linspace(p["f_min"], p["f_max"], n_masked)
    frequencies_masked = domain()[mask]
    assert np.linalg.norm(frequencies_masked - frequencies_expected_masked) < 1e-15


# def test_uniform_FD_noise_std(uniform_FD_params):
#     p = uniform_FD_params
#     domain = FrequencyDomain(**p)
#     expected = np.sqrt(p['window_factor']) / np.sqrt(4.0 * p['delta_f'])
#     assert np.abs(domain.noise_std - expected) < 1e-15


def test_FD_domain_dict(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    domain2 = build_domain(domain.domain_dict)
    assert domain.__dict__ == domain2.__dict__


def test_FD_update_data(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    n = int(p["f_max"] / p["delta_f"]) + 1
    nt = int((p["f_max"] - p["f_min"]) / p["delta_f"]) + 1
    # test that the sample frequencies are correct
    assert n == len(domain) == len(domain())
    assert nt == int(np.sum(domain.frequency_mask))
    assert np.all(domain() == np.linspace(0, p["f_max"], n))
    assert np.all(
        domain()[np.nonzero(domain.frequency_mask)]
        == np.linspace(p["f_min"], p["f_max"], nt)
    )
    # Test that data range adjustment works specifying different axis
    a = np.random.rand(10, n, 20)
    b = domain.update_data(a, axis=1)
    assert np.all(b[:, : domain.min_idx, :] == 0.0)
    assert np.all(b[:, domain.min_idx : n + 1, :] == a[:, domain.min_idx :, :])
    # Test that data range adjustment works as intended with nonzero low_value
    a = np.random.random(n)
    assert np.all(domain.update_data(a) == a * domain.frequency_mask)
    assert np.all(
        domain.update_data(a, low_value=5.0)[: domain.min_idx]
        == 5.0 * np.ones(round(p["f_min"] / p["delta_f"]))
    )


def test_FD_set_new_range(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    # test that ValueErrors are raised for infeasible inputs
    with pytest.raises(ValueError):
        domain.set_new_range(p["f_max"] + 10, None)
    with pytest.raises(ValueError):
        domain.set_new_range(p["f_min"] - 10, None)
    with pytest.raises(ValueError):
        domain.set_new_range(None, p["f_max"] + 10)
    with pytest.raises(ValueError):
        domain.set_new_range(None, p["f_min"] - 10)
    with pytest.raises(ValueError):
        domain.set_new_range(p["f_min"] + 10, p["f_min"] + 5)
    # test that setting new frequency range works as intended
    f_min_new, f_max_new = 40, 800
    n = int(f_max_new / p["delta_f"]) + 1
    domain.set_new_range(f_min_new, f_max_new)
    assert n == len(domain) == len(domain())
    assert np.all(domain() == np.linspace(0, f_max_new, n))
    mask = np.ones(n)
    mask[: round(f_min_new / p["delta_f"])] = 0.0
    assert np.all(domain.frequency_mask == mask)
    # Test that data range adjustment works as intended when setting new fmin, fmax
    a = np.random.random(n)
    assert np.all(domain.update_data(a) == a * domain.frequency_mask)
    assert np.all(
        domain.update_data(a, low_value=5.0)[: domain.min_idx]
        == 5.0 * np.ones(round(f_min_new / p["delta_f"]))
    )


def test_FD_time_translation(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    # test normal time translation
    dt = 1e-3
    a = np.sin(np.outer(np.arange(3) + 1, domain()) / 100) + 1j * np.cos(
        np.outer(np.arange(3) + 1, domain()) / 100
    )
    a_dt = a * np.exp(-2j * np.pi * domain() * dt)
    b_dt = domain.time_translate_data(a, dt)
    assert np.allclose(a_dt, b_dt)
    assert not np.allclose(a_dt, domain.time_translate_data(a, 1.01 * dt))
    # test that time translation can be undone
    assert np.allclose(a, domain.time_translate_data(a_dt, -dt))
    assert not np.allclose(a, domain.time_translate_data(a_dt, -1.01 * dt))


def test_FD_time_translation_torch(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    batch_size = 5
    num_detectors = 2
    dt = torch.randn(batch_size, num_detectors, dtype=torch.float32)
    data = torch.empty(
        (batch_size, num_detectors, 3, len(domain) - domain.min_idx),
        dtype=torch.float32,
    )
    constant_value = 2.0
    f = domain.sample_frequencies_torch[domain.min_idx :]
    for i in range(batch_size):
        for j in range(num_detectors):
            signal = torch.exp(-2j * np.pi * dt[i, j] * f)
            data[i, j, 0, :] = signal.real
            data[i, j, 1, :] = signal.imag
            data[i, j, 2, :] = constant_value
    result = domain.time_translate_data(data, -dt)
    assert result.dtype == torch.float32
    assert result.shape == data.shape
    assert torch.allclose(result[..., 0, :], torch.tensor(1.0), atol=1e-6)
    # Tolerance of 1e-2 is required in the imaginary part, likely because we are
    # checking that it *vanishes*. This is a consequence of single-precision floats.
    # TODO: Is there a way to improve on this?
    assert torch.allclose(result[..., 1, :], torch.tensor(0.0), atol=1e-2)
    assert torch.allclose(result[..., 2, :], torch.tensor(constant_value))


def test_FD_caching(uniform_FD_params):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    domain_ref = FrequencyDomain(**p)

    assert np.all(domain() == domain_ref())
    # we now modify domain._f_max by hand, which should not be done, as this
    # does not update the cached properties
    domain._f_max = 50
    assert np.all(domain() == domain_ref())
    domain._reset_caches()
    # after clearing the cache, the __call__ method should return the correct
    # result
    assert np.all(domain() != domain_ref())


def test_FD_window_factor(uniform_FD_params, window_setup):
    p = uniform_FD_params
    domain = FrequencyDomain(**p)
    _, window_factor = window_setup
    assert window_factor == 0.9374713897717841
    # check that window_factor is initially None
    assert domain.window_factor is None
    # set new window_factor
    domain.window_factor = window_factor
    assert domain._window_factor == domain.window_factor == window_factor
    noise_std = domain.noise_std
    assert noise_std == np.sqrt(domain.window_factor) / np.sqrt(4 * domain.delta_f)
    window_factor = 1
    # now set new window factor correctly via the setter and check that
    # noise_std is updated as intended since the cache is cleared
    domain.window_factor = window_factor
    assert domain._window_factor == domain.window_factor == window_factor
    assert domain.noise_std != noise_std
    assert domain.noise_std == np.sqrt(domain.window_factor) / np.sqrt(
        4 * domain.delta_f
    )


# def test_FD_truncation_old(uniform_FD_params):
#     p = uniform_FD_params
#     domain = FrequencyDomain(**p)
#     N = len(domain)
#     domain.initialize_truncation((40,1024))
#     assert domain._truncated_sample_frequencies[0] == 40
#     assert domain._truncated_sample_frequencies[-1] == 1024
#     # test that array can be truncated with automatic axis selection
#     a = np.random.rand(3, N, 4, 2)
#     a_truncated = domain.truncate_data(a)
#     assert a_truncated.shape[1] == domain._truncation_num_bins
#     assert a_truncated.shape[0] == a.shape[0]
#     assert a_truncated.shape[2:] == a.shape[2:]
#     assert np.all(a[:,domain._truncation_idx_lower:domain._truncation_idx_upper]
#                   == a_truncated)
#     assert not np.all(a[:,0:domain._truncation_num_bins] == a_truncated)
#     # test that axis can be selected manually
#     a_truncated_2 = domain.truncate_data(a, axis=1)
#     assert np.all(a_truncated == a_truncated_2)
#     # test that errors are raised as intended
#     with pytest.raises(ValueError):
#         domain.truncate_data(a, axis=0)
#     with pytest.raises(ValueError):
#         domain.truncate_data(np.zeros((10, N-1, N+1)))
#     with pytest.raises(ValueError):
#         domain.truncate_data(np.zeros((10, N, N)))
#     # test that manual axis selection works in the above case
#     assert domain.truncate_data(np.zeros((10, N, N)), axis=1).shape == \
#            (10, domain._truncation_num_bins, N)

# def test_FD_time_translation_old(uniform_FD_params):
#     p = uniform_FD_params
#     domain = FrequencyDomain(**p)
#     domain.initialize_truncation((40,1024))
#     dt = 1e-3
#     data = np.sin(np.outer(np.arange(3)+1, domain())/100) + \
#            1j * np.cos(np.outer(np.arange(3)+1, domain())/100)
#     data_translated = domain.time_translate_data(data, dt)
#     # check that time translation has correct inverse
#     assert not np.allclose(data_translated, data)
#     assert np.allclose(domain.time_translate_data(data_translated, -dt), data)
#     # check that truncation and time translation commutes
#     data_truncated = domain.truncate_data(data)
#     data_truncated_translated = domain.time_translate_data(data_truncated, dt)
#     data_translated_truncated = domain.truncate_data(data_translated)
#     assert np.all(data_translated_truncated == data_truncated_translated)
#     # import matplotlib.pyplot as plt
#     # plt.plot(data[0].real)
#     # plt.plot(data_translated[0].real)
#     # plt.plot(domain.time_translate_data(data_translated, -dt)[0].real)
#     # plt.show()


# def test_TD():
#     time_duration, sampling_rate = 4.0, 1.0 / 4096.0
#     domain = TimeDomain(time_duration, sampling_rate)
#     delta_t = 1.0 / sampling_rate
#     n = time_duration / delta_t
#     times_expected = np.arange(n) * delta_t
#     assert np.linalg.norm(domain() - times_expected) < 1e-15



from os.path import isfile

import numpy as np
from gwpy.timeseries import TimeSeries

from dingo.core.dataset import DingoDataset
from dingo.core.utils.misc import recursive_check_dicts_are_equal
from dingo.gw.data.data_download import download_raw_data
from dingo.gw.gwutils import get_window
from dingo.gw.domains import build_domain_from_model_metadata, FrequencyDomain


def load_raw_data(time_event, settings, event_dataset=None):
    """
    Load raw event data.

    * If event_dataset is provided and event data is saved in it, load and return the data
    * Else, event data is downloaded. If event_dataset is provided, the event data is
      additionally saved to the file.

    Parameters
    ----------
    time_event: float
        gps time of the events
    settings: dict
        dict with the settings
    event_dataset: str
        name of the event dataset file

    Returns
    -------

    """
    event = str(time_event)

    # first try to load the event data from the saved dataset
    if event_dataset is not None:

        if isfile(event_dataset):
            dataset = DingoDataset(file_name=event_dataset, data_keys=[event])
            if settings is not None:
                if not recursive_check_dicts_are_equal(settings, dataset.settings):
                    raise ValueError(
                        f"Settings {settings} don't match saved settings "
                        f"{dataset.settings}"
                    )
            data = vars(dataset)[event]
            print(f"Data for event at {event} found in {event_dataset}.")
            return data

    # if this did not work, download the data
    print(f"Downloading data for event at {event}.")
    data = download_raw_data(time_event, **settings)

    # optionally save this data to event_dataset
    if event_dataset is not None:
        dataset = DingoDataset(
            dictionary={event: data, "settings": settings}, data_keys=[event]
        )
        print(f"Saving data for event at {event} to {event_dataset}.")
        dataset.to_file(event_dataset, mode="a")

    return data


def parse_settings_for_raw_data(model_metadata, time_psd, time_buffer):
    domain_type = model_metadata["dataset_settings"]["domain"]["type"]

    if domain_type == "FrequencyDomain":
        data_settings = model_metadata["train_settings"]["data"]
        settings = {
            "window": data_settings["window"],
            "detectors": data_settings["detectors"],
            "time_segment": data_settings["window"]["T"],
            "time_psd": time_psd,
            "time_buffer": time_buffer,
            "f_s": data_settings["window"]["f_s"],
        }
    else:
        raise NotImplementedError(f"Unknown domain type {domain_type}")

    return settings


def data_to_domain(raw_data, settings_raw_data, domain, **kwargs):
    """

    Parameters
    ----------
    raw_data
    settings_raw_data
    model_metadata

    Returns
    -------
    data: dict
        dict with domain_data

    """

    if type(domain) == FrequencyDomain:
        window = get_window(kwargs["window"])
        data = {"waveform": {}, "asds": {}}
        # convert event strains to frequency domain
        for det, event_strain in raw_data["strain"].items():
            event_strain = TimeSeries(event_strain, dt=1 / settings_raw_data["f_s"])
            event_strain = event_strain.to_pycbc()
            event_strain = (event_strain * window).to_frequencyseries()
            event_strain = event_strain.cyclic_time_shift(
                settings_raw_data["time_buffer"]
            )
            event_strain = domain.update_data(np.array(event_strain))
            data["waveform"][det] = event_strain

        # convert psds to asds
        for det, psd in raw_data["psd"].items():
            asd = psd**0.5
            asd = domain.update_data(asd, low_value=1.0)
            data["asds"][det] = asd

        return data

    else:
        raise NotImplementedError(f"Unknown domain type {type(domain)}")


def get_event_data_and_domain(
    model_metadata,
    time_event,
    time_psd,
    time_buffer,
    event_dataset=None,
):
    # step 1: download raw event data
    settings_raw_data = parse_settings_for_raw_data(
        model_metadata, time_psd, time_buffer
    )
    raw_data = load_raw_data(
        time_event, settings=settings_raw_data, event_dataset=event_dataset
    )

    # step 2: prepare the data for the network domain
    domain = build_domain_from_model_metadata(model_metadata)
    event_data = data_to_domain(
        raw_data,
        settings_raw_data,
        domain,
        window=model_metadata["train_settings"]["data"]["window"],
    )

    return event_data, domain



import argparse

import h5py


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument("dataset", type=str)
    args = parser.parse_args()

    f = h5py.File(args.dataset, "r+")
    try:
        grp = f.create_group("svd")
        grp["V"] = f["svd_V"]
        del f["svd_V"]
    except ValueError:
        print("Dataset is already in correct format.")

    f.close()


if __name__ == "__main__":
    main()
